from napari import current_viewer
from magicgui import magicgui
from napari.utils.history import get_save_history, update_save_history 
import pathlib
import epicure.Utils as ut
from epicure.epicuring import EpiCure
import multiprocessing

"""
   Start EpiCure plugin
   Open the interface to select the movie and associated segmentation to process
"""

def start_epicure():
    hist = get_save_history()
    cdir = hist[0]
    viewer = current_viewer()
    Epic = EpiCure(viewer)
    caxis = None
    cval = 0
    ncpus = int(multiprocessing.cpu_count()*0.5)

    def set_visibility():
        """ Handle the visibility of the advanced parameters """
        get_files.output_dirname.visible = get_files.advanced_parameters.value
        get_files.show_other_chanels.visible = get_files.advanced_parameters.value
        get_files.process_frames_parallel.visible = get_files.advanced_parameters.value
        get_files.nbparallel_threads.visible = get_files.advanced_parameters.value
        get_files.junction_half_thickness.visible = get_files.advanced_parameters.value
        get_files.verbose_level.visible = get_files.advanced_parameters.value
        get_files.allow_gaps.visible = get_files.advanced_parameters.value
        get_files.show_scale_bar.visible = get_files.advanced_parameters.value
        get_files.epithelial_cells.visible = get_files.advanced_parameters.value

    def load_movie():
        """ Load and display the selected movie """
        start_time = ut.start_time()
        nonlocal caxis, cval
        image_file = get_files.image_file.value
        caxis, cval = Epic.load_movie(image_file)
        imgdir = ut.get_directory(image_file)
        get_files.segmentation_file.visible = True
        get_files.segmentation_file.value = pathlib.Path(imgdir)
        labname = Epic.suggest_segfile( get_files.output_dirname.value )
        Epic.set_names( get_files.output_dirname.value )
        if labname is not None:
            get_files.segmentation_file.value = pathlib.Path(labname)
            Epic.read_epicure_metadata()    
        if caxis is not None:
            get_files.junction_chanel.max = cval-1
            get_files.junction_chanel.visible = True
            set_chanel()
        get_files.scale_xy.value = Epic.epi_metadata["ScaleXY"]
        get_files.timeframe.value = Epic.epi_metadata["ScaleT"]
        get_files.unit_xy.value = Epic.epi_metadata["UnitXY"]
        get_files.unit_t.value = Epic.epi_metadata["UnitT"]
        get_files.scale_xy.visible = True
        get_files.unit_xy.visible = True
        get_files.timeframe.visible = True
        get_files.unit_t.visible = True
        get_files.segment_with_epyseg.visible = True
        get_files.allow_gaps.value = bool(Epic.epi_metadata["Allow gaps"])
        get_files.verbose_level.value = int(Epic.epi_metadata["Verbose"])
        get_files.call_button.enabled = True
        ut.show_duration(start_time, header="Movie loaded in ")

    def show_others():
        """ Display other chanels from the initial movie """
        for ochan in range(cval):
            ut.remove_layer(viewer, "MovieChannel_"+str(ochan))
        if get_files.show_other_chanels.value == True:
            Epic.add_other_chanels(int(get_files.junction_chanel.value), caxis)

    def set_chanel():
        """ Set the correct chanel that contains the junction signal """
        start_time = ut.start_time()
        Epic.set_chanel( int(get_files.junction_chanel.value), caxis )
        show_others()
        ut.show_duration(start_time, header="Movie chanel loaded in ")

    def launch_napari_epyseg():
        """ Open napari-epyseg plugin to segment the intensity channel movie """
        try:
            import napari_epyseg
            from napari_epyseg.start_epyseg import run_epyseg
        except:
            ut.show_error( "This option requires the plugin napari-epyseg that is missing.\nInstall it and restart" )
            return
        print("Running EpySeg with default parameters on the movie. To change the settings, use the napari-epyseg plugin outside of EpiCure or EpySeg module directly")
        parameters = {"tile_width":256, "tile_height":256, "overlap_width":32, "overlap_height":32, "model":"epyseg default(v2)", "norm_min":0, "norm_max":1}
        print(Epic.img.shape)
        ut.show_progress( viewer, True )
        segres = run_epyseg( Epic.img, parameters, prog=True )
        ut.show_progress( viewer, False )
        segname = str(get_files.image_file.value)+"_epyseg.tif"
        ut.writeTif( segres, segname, 1.0, "uint8", what="Epyseg results saved in " )
        get_files.segmentation_file.value = segname
        get_files.segment_with_epyseg.visible = False


    @magicgui(call_button="START CURE",
            junction_chanel={"widget_type": "Slider", "min":0, "max": 0},
            segment_with_epyseg = {"widget_type": "PushButton", "label": "Segment now with EpySeg"},
            scale_xy = {"widget_type": "LiteralEvalLineEdit"},
            timeframe = {"widget_type": "LiteralEvalLineEdit"},
            junction_half_thickness={"widget_type": "LiteralEvalLineEdit"},
            nbparallel_threads = {"widget_type": "LiteralEvalLineEdit"},
            verbose_level={"widget_type": "Slider", "min":0, "max": 3},
            )
    def get_files( 
                   image_file = pathlib.Path(cdir),
                   junction_chanel = 0,
                   segmentation_file = pathlib.Path(cdir),
                   segment_with_epyseg = False,
                   scale_xy = 1,
                   unit_xy = "um",
                   timeframe = 1,
                   unit_t = "min",
                   advanced_parameters = False,
                   show_other_chanels = True,
                   show_scale_bar = True,
                   allow_gaps = True,
                   epithelial_cells = True,
                   process_frames_parallel = False,
                   nbparallel_threads = ncpus,
                   junction_half_thickness = 1,
                   output_dirname = "epics",
                   verbose_level = 1,
                   ):
        
        print("Starting")
        imname, imdir, outdir = ut.extract_names( image_file, output_dirname )
        update_save_history(imdir)
        #ut.remove_widget(viewer, "Start EpiCure (epicure)")
        ut.remove_all_widgets( viewer )
        Epic.process_parallel = process_frames_parallel
        Epic.set_verbose( verbose_level )
        Epic.nparallel = nbparallel_threads
        #Epic.load_segmentation(segmentation_file)
        Epic.set_thickness( junction_half_thickness )
        Epic.set_scales(scale_xy, timeframe, unit_xy, unit_t)
        Epic.set_scalebar( show_scale_bar )
        Epic.set_gaps_option( allow_gaps )
        Epic.set_epithelia( epithelial_cells )
        Epic.go_epicure(outdir, segmentation_file)

    set_visibility()
    get_files.call_button.enabled = False
    get_files.segmentation_file.visible = False
    get_files.segment_with_epyseg.visible = False
    get_files.scale_xy.visible = False
    get_files.unit_xy.visible = False
    get_files.timeframe.visible = False
    get_files.unit_t.visible = False
    get_files.junction_chanel.visible = False
    get_files.advanced_parameters.clicked.connect(set_visibility)
    get_files.show_other_chanels.clicked.connect(show_others)
    get_files.image_file.changed.connect(load_movie)
    get_files.junction_chanel.changed.connect(set_chanel)
    get_files.segment_with_epyseg.clicked.connect( launch_napari_epyseg )
    return get_files




import numpy as np
import epicure.concatenate_movie as cm
import napari
import os
import epicure.epicuring as epi


def test_process_first_movie():
    main_dir = os.path.join(".", "data_test", "splitted")
    img = os.path.join( main_dir, "013-t1-t4.tif" )
    seg = os.path.join( main_dir, "013-t1-t4_epyseg.tif" )

    print("Open first movie")
    epic = epi.EpiCure()
    epic.load_movie( img )
    assert epic.img.shape == (4, 168, 219)
    assert epic.viewer is not None
    assert epic.nframes == 4

    print("Load epyseg segmentation")
    epic.go_epicure( "epics", seg )
    assert epic.tracked == 0
    assert epic.nlabels() == 494

    print("Remove border cells")
    epic.editing.remove_border()
    assert epic.nlabels() == 275
    print(epic.get_labels())

    print("Assign some cells to groups")
    epic.reset_groups()
    epic.cells_ingroup(17, "Bing")
    epic.cells_ingroup(42, "Bing")
    epic.cells_ingroup(142, "Bang")
    epic.cells_ingroup(143, "Bang")

    print("Do tracking with default parameters")
    epic.tracking.do_tracking()
    assert epic.nlabels() == 73

    print("Generate suspect list")
    epic.inspecting.inspect_tracks()
    assert len(epic.inspecting.events.data) == 4

    epic.save_epicures()
    assert os.path.exists(os.path.join( main_dir, "epics", "013-t1-t4_labels.tif") )
    assert os.path.exists(os.path.join( main_dir, "epics", "013-t1-t4_epidata.pkl") )

def test_process_second_movie():
    main_dir = os.path.join(".", "data_test", "splitted")
    img = os.path.join( main_dir, "013-t4-t6.tif" )
    seg = os.path.join( main_dir, "013-t4-t6_epyseg.tif" )
    
    print("Open second movie")
    epic = epi.EpiCure()
    epic.viewer = napari.Viewer(show=False)
    epic.load_movie( img )
    assert epic.img.shape == (3, 168, 219)
    assert epic.viewer is not None
    
    epic.go_epicure( "epics", seg )
    assert epic.tracked == 0
    assert epic.nlabels() == 371
    
    epic.tracking.do_tracking()
    epic.reset_groups()
    epic.cells_ingroup(21, "Bing")
    epic.cells_ingroup(22, "BAM")
    epic.save_epicures()
    assert os.path.exists(os.path.join( main_dir, "epics", "013-t4-t6_labels.tif") )
    assert os.path.exists(os.path.join( main_dir, "epics", "013-t4-t6_epidata.pkl") )


def test_merge_movies():
    main_dir = os.path.join(".", "data_test", "splitted")
    img1 = os.path.join( main_dir, "013-t1-t4.tif" )
    lab1 = os.path.join( main_dir, "epics", "013-t1-t4_labels.tif" )
    img2 = os.path.join( main_dir, "013-t4-t6.tif" )
    lab2 = os.path.join( main_dir, "epics", "013-t4-t6_labels.tif" )

    cm.merge_epicures( img1, lab1, img2, lab2, "together.tif")
    assert os.path.exists(os.path.join( main_dir, "together.tif") )
    assert os.path.exists(os.path.join( main_dir, "epics", "together_labels.tif") )
    assert os.path.exists(os.path.join( main_dir, "epics", "together_epidata.pkl") )
    
    img = os.path.join( main_dir, "together.tif" )
    seg = os.path.join( main_dir, "epics", "together_labels.tif" )
    epic = epi.EpiCure()
    epic.viewer = napari.Viewer(show=False)
    epic.load_movie( img )
    assert epic.img.shape == (6, 168, 219)

    epic.go_epicure( "epics", seg )
    assert len(epic.groups.keys()) == 3
    assert 142 in epic.groups["Bang"]
    assert 23 in epic.groups["Bing"]
    
    assert epic.tracked == 1
    epic.outputing.output_mode.setCurrentIndex(1)
    epic.outputing.show_trackfeature_table()
    assert "Label" in epic.outputing.trackTable.get_features_list()
    assert "TrackDuration" in epic.outputing.trackTable.get_features_list()
    headers = [epic.outputing.trackTable.wid_table.horizontalHeaderItem(ind).text() for ind in range(epic.outputing.trackTable.wid_table.columnCount()) ]
    durind = headers.index("TrackDuration")
    assert int(epic.outputing.trackTable.wid_table.item(20, durind).text()) == 6



#process_first_movie()
#process_second_movie()

#test_merge_movies()



import numpy as np
import os
import epicure.epicuring as epi
import epicure.Utils as ut
from unittest.mock import Mock
import napari
from vispy import keys


def test_output_selected():
    test_img = os.path.join(".", "data_test", "area3_t100-101.tif")
    test_seg = os.path.join(".", "data_test", "area3_epyseg-t100-101.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    resaxis, resval = epic.load_movie(test_img)
    epic.set_chanel(1, 1)
    assert epic.viewer is not None
    epic.go_epicure("test_epics", test_seg)

    output = epic.outputing
    assert output is not None
    sel = output.get_selection_name()
    assert sel == ""
    output.output_mode.setCurrentText("All cells")
    sel = output.get_selection_name()
    assert sel == ""
    output.output_mode.setCurrentText("Only selected cell")
    sel = output.get_selection_name()
    assert sel == "_cell_1"
    roi_file = os.path.join(".", "data_test", "test_epics", "area3_t100-101_rois_cell_1.zip")
    if os.path.exists(roi_file):
        os.remove(roi_file)
    ## TO UPDATE WITH NEW VERSION
    #output.roi_out()
    #assert os.path.exists(roi_file)

#test_output_selected()



"""
Tracking with laptrack package

Inspired from example https://github.com/yfukai/laptrack/blob/main/docs/examples/cell_segmentation.ipynb
"""

import numpy as np
import pandas as pd

import laptrack
from laptrack import LapTrack
import epicure.Utils as ut
from packaging.version import Version
    

def squared_difference(a1, a2):
    """ Squared difference, normalized """
    return ((a1-a2)**2)/(max(a1, a2)**2)

def squared_distance(x1, y1, x2, y2):
    """ Return squared distance """
    return (x1-x2)**2 + (y1-y2)**2


class LaptrackCentroids():

    def __init__(self, track, epic):
        self.max_distance = 15
        self.splitting_cost = 1
        self.merging_cost = 1
        self.penal_area = 0
        self.penal_solidity = 0
        self.track = track
        self.epicure = epic
        self.inspecting = False
        self.suggesting = False
        self.region_properties = ["label", "frame", "centroid-0", "centroid-1", "area", "solidity"]
        ## to handle difference in laptrack versions
        self.version_over = Version(laptrack.__version__) >= Version("0.17.0")

        if self.version_over:
            self.lt = LapTrack( metric=self.tracking_metric,
                                cutoff=1, 
                                splitting_metric=self.tracking_nofeat_metric, 
                                splitting_cutoff=self.splitting_cost, 
                                merging_metric=self.tracking_nofeat_metric,
                                merging_cutoff=self.merging_cost, )
        else:
            self.lt = LapTrack( track_dist_metric=self.tracking_metric,
                                track_cost_cutoff=1, 
                                splitting_dist_metric=self.tracking_nofeat_metric, 
                                splitting_cost_cutoff=self.splitting_cost, 
                                merging_dist_metric=self.tracking_nofeat_metric,
                                merging_cost_cutoff=self.merging_cost, )
        
    def set_region_properties(self, with_extra=False):
        """ define the region properties used in the tracker """
        if with_extra:
            self.region_properties = ["label", "frame", "centroid-0", "centroid-1", "area", "solidity"]
        else:
            self.region_properties = ["label", "frame", "centroid-0", "centroid-1" ]

    def twoframes_track(self, df, labels):
        """ Do track between two frames, only track label """
        #start_time = time.time()
        track_df, split_df, merge_df = self.perform_track( df )
        #show_info("Performed in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        
        track_ids = [None]*len(labels)
        ## look for track id associated with label
        frame_df = track_df[track_df["frame"] == 1]
        label_to_tid = {int(row["label"]): int(row["track_id"]) for i, row in frame_df.iterrows()}
        for i, curlabel in enumerate(labels):
            track_ids[i] = label_to_tid.get(curlabel, None)

        tracklabels = [None]*len(labels)
        ## look for cell label associated with track_id in the first frame, if any
        frame_df = track_df[track_df["frame"]==0]
        frame_df = frame_df[frame_df["track_id"].isin(track_ids)]
        label_to_tid = {int(row["track_id"]): int(row["label"]) for i, row in frame_df.iterrows()}
        for i, tid in enumerate(track_ids):
            tracklabels[i] = label_to_tid.get(tid, None)
        #show_info("Finished in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        return tracklabels

    def tracking_metric(self, c1, c2):
        """ Tracking cost metric: distance + feature penalties """
        dist = squared_distance(c1[2], c1[3], c2[2], c2[3])
        if dist > self.max_distance**2:
            return dist
        ## normalize all distances to combine them
        dist = dist / (self.max_distance**2)
        if self.penal_area > 0:
            dist += self.penal_area * squared_difference(c1[4], c2[4])
        if self.penal_solidity > 0:
            dist += self.penal_solidity * squared_difference(c1[5], c2[5])
        return dist
    
    def tracking_nofeat_metric(self, c1, c2):
        """ Tracking cost metric: distance, no penalties """
        dist = squared_distance(c1[2], c1[3], c2[2], c2[3])
        if dist > self.max_distance**2:
            return dist
        ## normalize all distances to combine them
        dist = dist / (self.max_distance**2)
        return dist

    def perform_track(self, regionprops_df):
        """ Do tracking with laptrack module """
        split_cost = False
        merg_cost = False
        #from laptrack import ParallelBackend
        #self.lt.parallel_backend = ParallelBackend.ray
        if self.splitting_cost > 0:
            split_cost = self.splitting_cost**2
        if self.version_over:
            self.lt.splitting_cutoff = split_cost
        else:
            self.lt.splitting_cost_cutoff = split_cost
        if self.merging_cost > 0:
            merg_cost = self.merging_cost**2
        if self.version_over:
            self.lt.merging_cutoff = merg_cost
        else:
            self.lt.merging_cost_cutoff = merg_cost
        track_df, split_df, merge_df = self.lt.predict_dataframe(
            regionprops_df,
            coordinate_cols=self.region_properties,
            only_coordinate_cols=False,
        )    
        track_df = track_df.reset_index()
        return track_df, split_df, merge_df

    def track_centroids(self, regionprops_df):
        """ Track cells based on their centroids positions + features penalties """
        ut.napari_info("Starting tracking with LapTrack centroids metrics...")
        return self.perform_track( regionprops_df )

    def inspect_oneframe(self, graph, trackdf):
        for track in np.unique(trackdf["track_id"]):
            tr = trackdf[trackdf["track_id"] == track]
            ## track is only on one frame, suspect
            if len(np.unique(tr["frame"])) == 1:
                # trackid + 1 as trackid starts as 0
                pos = (tr.iloc[0]["frame"], int(tr.iloc[0]["centroid-0"]), int(tr.iloc[0]["centroid-1"]))
                self.epicure.inspecting.add_event( pos, track+1, "tracking" )
                if self.track.suggesting:
                    if track in graph.keys():
                        sisters = []
                        refval = graph[track][0]
                        for key, val in graph.items():
                            if val[0] == refval:
                                sisters.append( key )
                        if len(sisters) == 2:
                            for sis in sisters:
                                self.epicure.add_suggestion( sis+1, refval+1 )




import numpy as np
import os
import epicure.epicuring as epi

def test_track_methods():

    test_img = os.path.join(".", "data_test", "003_crop.tif")
    test_seg = os.path.join(".", "data_test", "003_crop_epyseg.tif")

    ## load and initialize
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)
    
    assert epic.tracked == 0
    assert epic.nlabels() == 2842

    track = epic.tracking
    # default tracking
    track.do_tracking()
    assert epic.nlabels() == 347
    assert epic.nlabels() == track.nb_tracks()
    assert track.graph is not None
    assert 1170 in track.graph.keys()
    assert track.graph[855] == [80, 83]
    assert not track.check_gap()

    track.track_choice.setCurrentText("Laptrack-Overlaps")
    track.do_tracking()
    assert epic.nlabels() == track.nb_tracks()
    assert track.graph is not None
    assert epic.nlabels() == 309
    assert not track.check_gap()

    track.reset_tracks()
    assert epic.nlabels() == 309
    assert epic.nlabels() == track.nb_tracks()

    ## check one track validity
    track_id = track.get_track_list()[20]
    assert track_id == 21
    assert track.get_first_frame( track_id ) == 0
    feats = track.measure_track_features( track_id )
    assert feats["TrackDuration"] == 8

    last = track.get_last_frame(track_id)
    track.remove_one_frame( track_id, last )
    newlast = track.get_last_frame(track_id)
    assert last == newlast+1

    ## create a gap in the middle of the track, then fix it (split the track)
    midle = track.get_first_frame(track_id) + 2
    track.remove_one_frame( track_id, midle )
    gaped = track.check_gap()
    ## gaps are allowed now
    assert len(gaped) > 0
    epic.handle_gaps( None )
    assert track.nb_tracks() == 310

#track_methods()
#test_track_methods()


import numpy as np
import os
import epicure.epicuring as epi
import epicure.Utils as ut
from unittest.mock import Mock
import napari
from vispy import keys

def test_epicuring_bindings():
    test_img = os.path.join(".", "data_test", "003_crop.tif")
    test_seg = os.path.join(".", "data_test", "003_crop_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)
    assert epic.viewer is not None
    ut.set_active_layer(epic.viewer, "Segmentation")
    view = epic.viewer.window.qt_viewer
    view.canvas.events.key_press(key=keys.Key("b"), modifiers=[])
    assert epic.viewer.layers["Segmentation"].visible == False
    view.canvas.events.key_press(key=keys.Key("b"), modifiers=[])
    assert epic.viewer.layers["Segmentation"].visible == True
    view.canvas.events.key_press(key=keys.Key("v"), modifiers=[])
    assert epic.viewer.layers["Movie"].visible == False
    view.canvas.events.key_press(key=keys.Key("v"), modifiers=[])
    assert epic.viewer.layers["Movie"].visible == True
    view.canvas.events.key_press(key=keys.Key("c"), modifiers=[])
    assert epic.viewer.layers["Movie"].visible == True
    assert epic.viewer.layers["Segmentation"].visible == False
    view.canvas.events.key_press(key=keys.Key("g"), modifiers=[])
    assert "EpicGrid" in epic.viewer.layers
    
def test_merge():
    test_img = os.path.join(".", "data_test", "003_crop.tif")
    test_seg = os.path.join(".", "data_test", "003_crop_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.load_segmentation(test_seg)
    epic.go_epicure("test_epics", test_seg)
    assert epic.viewer is not None

    segedit = epic.editing
    assert segedit is not None
    val = epic.seglayer.data[0,93,167]
    assert val == 111
    ## check that not touching labels do not merge
    segedit.merge_labels( 0, 111, 134 )
    val = epic.seglayer.data[0,93,167]
    assert val == 111
    segedit.merge_labels( 0, 111, 102 )
    val = epic.seglayer.data[0,93,167]
    ## new label because no propagation
    assert val == 2844
    

def test_group():
    test_img = os.path.join(".", "data_test", "003_crop.tif")
    test_seg = os.path.join(".", "data_test", "003_crop_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)
    imname, imdir, outdir = ut.extract_names( test_img, "epics" )
    assert epic.viewer is not None
    layer = epic.viewer.layers["Segmentation"]
    
    epic.cells_ingroup( 111, "Test" )
    assert "Test" in epic.groups

    segedit = epic.editing
    event = napari.utils.events.Event("mouse_press") #viewer.window.qt_viewer.canvas.events.mouse_press(pos=(1, 2), modifiers=('Shift'), button=1)
    event.position = [0,10,20]
    event.view_direction = None
    event.dims_displayed = [0,1, 1] 
    assert "GroupTest" not in epic.groups
    segedit.group_choice.setCurrentText("GroupTest")
    segedit.add_cell_to_group(event)
    assert "GroupTest" in epic.groups
    


